%% Analysis of LTV Systems
% This file assesses the nominal induced L2 gain of an LTV system. The
% system is assumed to have a rational (LFT) dependence on time. The gain
% is calculated using the exact finite-horizon condition via the gridding 
% approximation.  This is compared against an infinite-horizon gain bound 
% computed using an IQC approach formulated with the swapping lemma.

%% Construct Time Varying Model 
% System has one state with pole at -p(t) where:
%   p(t) = m22 + (m12*m21*t)/(1-m11 t)
% Specify p0 = p(0), pinf = p(inf), and tstar := time at which
% p(tstar) = p0 + 0.9*(pinf-p0). Solve for the LFT entries:
%   m22=p0
%   m11 = -9/tstar
%   m12*m21 = -m11*(pinf-m22)
%

% Select pole initial value, final value, and time to 90%
p0 = 1;     pinf = 2;  tstar = 10;

% Construct LFT for p(t)
m22=p0;
m11 = -9/tstar;
fac = -m11*(pinf-m22);
if fac>=0
    m12 = sqrt(fac); m21 = m12;
else
    m12 = sqrt(abs(fac)); m21 = -m12;
end
M = [m11 m12; m21 m22];

% LFT Form: Glft = lft(t,Plft)
b = 1;
c = 5;
N = ss(0,[-1 b],[1;c],0);
Plft = lft(M,N,1,1);

%% Finite Horizon Nominal Gain
% This portion of the code uses the LTVTools toolbox

% Gridded Form
T=6; 
t = linspace(0,T,1e3); 
p = M(2,2) + M(2,1)*M(1,2)*t./(1-M(1,1)*t);
a = tvmat( -p, t);
Gg = tvss(a,b,c,0);

% Finite Horizon Analysis
Tdata = [0.01 0.1 0.5 1:T];
Nfh = numel(Tdata);
gfh1 = zeros(Nfh,1);
for i=1:Nfh
    gg = tvnorm(Gg,[0,Tdata(i)]);
    gfh1(i) = gg(1);
end

%% Infinite Horizon Nominal Gain - Computed with CVX
    
% IQC Filter
% Set nH=1 to obtain results for static multipliers H(s)=1.
nH = 8; 
H0 = ss(-2,1,1,0);
H = ss(1);
for i=2:nH
    H = [1; H0*H];
end

% Uncomment to add one more pole to filter
% H = [H; ss(-10,1,1,0)]; nH = nH+1; 

[ah,bh,ch,dh] = ssdata(H);
nx = size(ah,1);
Hb = ss(ah,bh,eye(nx),0);
Hc = ss(ah,eye(nx),ch,0);

% The code below first constructs a nonminimal realization for Psi and
% then uses minreal to remove unobservable/uncontrollable modes. It is
% preferable to directly construct a minimal realization.
%
%Psi = [H zeros(nH,1); Hc*Hb H];
%Psi= minreal(Psi);

% Minimial Realization for Psi
LL = blkdiag(eye(nH), ss(ah,eye(nx),ch,0));
CC = [ch zeros(nH,1); eye(nx) bh];
RR = blkdiag( ss(ah,bh,eye(nx),0),1);
Psi = LL*CC*RR + blkdiag(dh,dh);

% Augmented Plant
systemnames = 'Psi Plft';
inputvar = '[w; d]';
outputvar = '[Psi; Plft(2)]';
input_to_Psi = '[Plft(1); w]';
input_to_Plft = '[w; d]';
Paug = sysic; 

[aa,ba,ca,da] = ssdata(Paug);
na = size(aa,1);

Riqc = [ca(1:2*nH,:) da(1:2*nH,:)];
Rperf = [ca(2*nH+1:end,:) da(2*nH+1:end,:); zeros(1,na) 0 1];

% IQC LMI - CVX
% cvx_solver('SeDuMi');
% cvx_solver('SDPT3');
cvx_begin sdp quiet
    % Create variables
    variable P(na,na) symmetric;
    variable X(nH,nH) symmetric;
    if nH>1
        variable Y(nH,nH) skew_symmetric;
    end
    variable gsq;
        
    % Create KYP LMI
    if nH>1
        [aa'*P+P*aa P*ba; ba'*P zeros(2)] + ...
            Riqc'*[zeros(nH) X; X zeros(nH)]*Riqc + ...
            Riqc'*[zeros(nH) Y; Y' zeros(nH)]*Riqc + ...
            Rperf'*[1 0; 0 -gsq]*Rperf <= 0;       
    else
        [aa'*P+P*aa P*ba; ba'*P zeros(2)] + ...
            Riqc'*[zeros(nH) X; X zeros(nH)]*Riqc + ...
            Rperf'*[1 0; 0 -gsq]*Rperf <= 0;       
    end
    
    % Restrict matrix vars to be PSD
    P >= 0;
    X >= 0;
        
    % Objective function:  min trace(P)
    minimize( gsq );
cvx_end
gopt1 = sqrt(gsq);

%% Infinite Horizon Nominal Gain - Computed with LMILab
setlmis([])
[P,NDEC,PDEC] = lmivar(1,[na,1]);
[X,NDEC,XDEC] = lmivar(1,[nH,1]);
XX = lmivar(3,[zeros(nH) XDEC; XDEC zeros(nH)]);
if nH>1
    [Y,NDEC,YDEC] = lmivar(3,skewdec(nH,NDEC));
    YY = lmivar(3,[zeros(nH) YDEC; YDEC' zeros(nH)]);
end
[gsq,NDEC,gsqDEC] = lmivar(3,NDEC+1);

lmiterm([1 1 1 P],[eye(na); zeros(2,na)],[aa ba],'s');
lmiterm([1 1 1 XX],Riqc',Riqc);
if nH>1
    lmiterm([1 1 1 YY],Riqc',Riqc);
end
lmiterm([1 1 1 0],Rperf(1,:)'*Rperf(1,:));
lmiterm([-1 1 1 gsq],Rperf(2,:)',Rperf(2,:));

lmiterm([-2 1 1 X],1,1);
lmiterm([-3 1 1 P],1,1);

lmisys = getlmis;

cobj = zeros(NDEC,1);
cobj(end) = 1;

opt = [0 0 0 0 1];
[copt,XOPT] = mincx(lmisys,cobj,opt);
gopt2 = sqrt(copt);

% Print Results
[gfh1(end) gopt1 gopt2 gopt1/gfh1(end)]

%% Finite Horizon Nominal Gain - Using LFT IQC
gfh2 = zeros(Nfh,1);
tic
for i=1:Nfh
    cvx_begin sdp quiet
        % Create variables
        variable P(na,na) symmetric;
        variable X(nH,nH) symmetric;
        if nH>1
            variable Y(nH,nH) skew_symmetric;
        end
        variable gsq;
        variable Z(nH,nH) symmetric;

        % Create KYP LMI
        if nH>1
            [aa'*P+P*aa P*ba; ba'*P zeros(2)] + ...
                Riqc'*[zeros(nH) X; X zeros(nH)]*Riqc + ...
                Riqc'*[zeros(nH) Y; Y' zeros(nH)]*Riqc + ...
                Riqc'*[Tdata(i)^2*Z zeros(nH); zeros(nH) -Z]*Riqc + ...
                Rperf'*[1 0; 0 -gsq]*Rperf <= 0;       
        else
            [aa'*P+P*aa P*ba; ba'*P zeros(2)] + ...
                Riqc'*[zeros(nH) X; X zeros(nH)]*Riqc + ...
                Riqc'*[Tdata(i)^2*Z zeros(nH); zeros(nH) -Z]*Riqc + ...
                Rperf'*[1 0; 0 -gsq]*Rperf <= 0;                   
        end
        % Restrict matrix vars to be PSD
        P >= 0;
        X >= 0;
        Z >= 0;

        % Objective function:  min trace(P)
        minimize( gsq );
    cvx_end
    gfh2(i) = sqrt(gsq);
end
toc

%% Plot Results
figure(1);
plot(Tdata,gfh1,'b');
xlabel('Time Horizon, sec');
ylabel('L2 Gain')
hold on;
plot(Tdata,gfh2,'r-.');
plot(Tdata,gfh1,'bx');
plot(Tdata,gfh2,'ro');
legend('Riccati','IQC','Location','Best');
hold off
grid on;
% garyfyFigure