%% Analysis of LTV Systems
% This file assesses the robust induced L2 gain of an uncertain LTV system. 
% The system is assumed to have a rational (LFT) dependence on time. The
% analysis uses an IQC for the operator Delta_t:=t*I formulated with the 
% swapping lemma.

%% Construct Time Varying Model 
% System is second order with damping ratio given by:
%   z(t) = m22 + (m12*m21*t)/(1-m11 t)
% Specify z0 = z(0), zinf = z(inf), and tstar := time at which
% z(tstar) = z0 + 0.9*(zinf-z0). Solve for the LFT entries:
%   m22=z0
%   m11 = -9/tstar
%   m12*m21 = -m11*(zinf-m22)

% Select initial value, final value, and time to 90% for z(t)
z0 = 0.2;     zinf = 0.9;  tstar = 10;

% Construct LFT for z(t)
m22=z0;
m11 = -9/tstar;
fac = -m11*(zinf-m22);
if fac>=0
    m12 = sqrt(fac); m21 = m12;
else
    m12 = sqrt(abs(fac)); m21 = -m12;
end
M = [m11 m12; m21 m22];

DeltaT = ureal('DeltaT',0);
z = lft(DeltaT,M);

% Second Order System
A = [0 1; -4 -2*z];
B = [0;1];
C = [4 0];
D = 0;
Pltv = ss(A,B,C,D);

%% Sensitivity Function for Classical Feedback System

% Proportional Control
K = 2;

% Form LFT for uncertain LTV system with Delta = diag(Deltat,DeltaU)
DeltaU = ultidyn('DeltaU',[1 1]);
Sltv = feedback(1, Pltv*(1+DeltaU)*K);
[Glti,Del] = lftdata(Sltv);

%% Finite Horizon Nominal Gain
% This portion of the code uses the LTVTools toolbox

% Gridded Form
T=10; 
t = linspace(0,T,1e3); 
zg = M(2,2) + M(2,1)*M(1,2)*t./(1-M(1,1)*t);
zg = tvmat(zg,t);
Ag = [0 1; -4 -2*zg];
Gg = tvss(Ag,B,C,D);

% Nominal Gridded 
Sg = feedback(1,Gg*K);

% Finite Horizon Analysis
Tdata = [0.01 0.1 0.5 1:T];
Nfh = numel(Tdata);
gfh1 = zeros(Nfh,1);
for i=1:Nfh
    gg = tvnorm(Sg,[0,Tdata(i)]);
    gfh1(i) = gg(1);
end
[Tdata(:) gfh1]'


%% Infinite Horizon Nominal Gain - Computed with CVX
    
% IQC Filter
nH = 8;
H0 = ss(-2,1,1,0);
H = ss(1);
for i=2:nH
    H = [1; H0*H];
end
[ah,bh,ch,dh] = ssdata(H);
nx = size(ah,1);
Hb = ss(ah,bh,eye(nx),0);
Hc = ss(ah,eye(nx),ch,0);

% Minimal Realization for Psi1
LL = blkdiag(eye(nH), ss(ah,eye(nx),ch,0));
CC = [ch zeros(nH,1); eye(nx) bh];
RR = blkdiag( ss(ah,bh,eye(nx),0),1);
Psi1 = LL*CC*RR + blkdiag(dh,dh);

% Static IQC for DeltaU is (Psi2,M2) where M2 is given below
Psi2 = ss(eye(2));

% Augmented Plant
systemnames = 'Psi1 Psi2 Glti';
inputvar = '[w1; w2; d]';
outputvar = '[Psi1; Psi2; Glti(3)]';
input_to_Psi1 = '[Glti(1); w1]';
input_to_Psi2 = '[Glti(2); w2]';
input_to_Glti = '[w1; w2; d]';
Paug = sysic; 

[aa,ba,ca,da] = ssdata(Paug);
na = size(aa,1);

Riqc1 = [ca(1:2*nH,:) da(1:2*nH,:)];
Riqc2 = [ca((2*nH+1):(2*nH+2),:) da((2*nH+1):(2*nH+2),:)];
Rperf = [ca(2*nH+3:end,:) da(2*nH+3:end,:); zeros(1,na) 0 0 1];

% Compute Gains for Several Uncertainty Bounds
DeltaUBnd = [0:0.05:0.55]';
Ndel = numel(DeltaUBnd);
giqc = zeros(Ndel,1);


tic
for i = 1:Ndel
    M2 = [DeltaUBnd(i)^2 0; 0 -1];
    
    % IQC LMI - CVX
    % cvx_solver('SeDuMi');
    % cvx_solver('SDPT3');
    cvx_begin sdp quiet
        % Create variables
        variable P(na,na) symmetric;
        variable X(nH,nH) symmetric;
        variable Y(nH,nH) skew_symmetric;
        variable gsq;
        variable lam;

        % Create KYP LMI
        [aa'*P+P*aa P*ba; ba'*P zeros(3)] + ...
            Riqc1'*[zeros(nH) X; X zeros(nH)]*Riqc1 + ...
            Riqc1'*[zeros(nH) Y; Y' zeros(nH)]*Riqc1 + ...
            lam*Riqc2'*M2*Riqc2 + ...
            Rperf'*[1 0; 0 -gsq]*Rperf <= 0;

        % Restrict matrix vars to be PSD
        P >= 0;
        X >= 0;
        lam >= 0;

        % Objective function:  min trace(P)
        minimize( gsq );
    cvx_end
    giqc(i) = sqrt(gsq);
    [i DeltaUBnd(i) giqc(i)]
end
toc

%% Plot Results
figure(1);
plot(DeltaUBnd,giqc,'b',DeltaUBnd,giqc,'bx',0,gfh1(end),'ro');
axis([0 DeltaUBnd(end) 0 max(giqc)]);
grid;
xlabel('Uncertainty Bound');
ylabel('Induced L2 Gain');
garyfyFigure